package ru.dbotthepony.kstarbound

import com.google.gson.Gson
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.internal.bind.JsonTreeReader
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.api.IStarboundFile
import ru.dbotthepony.kstarbound.defs.player.RecipeDefinition
import ru.dbotthepony.kstarbound.util.KOptional
import java.util.Collections
import java.util.LinkedList
import java.util.concurrent.Callable
import java.util.concurrent.ExecutorService
import java.util.concurrent.ForkJoinPool
import java.util.concurrent.ForkJoinTask
import java.util.concurrent.Future

object RecipeRegistry {
	private val LOGGER = LogManager.getLogger()

	private val recipesInternal = ArrayList<RegistryObject<RecipeDefinition>>()
	private val group2recipesInternal = Object2ObjectOpenHashMap<String, LinkedList<RegistryObject<RecipeDefinition>>>()
	private val group2recipesBacking = Object2ObjectOpenHashMap<String, List<RegistryObject<RecipeDefinition>>>()
	private val output2recipesInternal = Object2ObjectOpenHashMap<String, LinkedList<RegistryObject<RecipeDefinition>>>()
	private val output2recipesBacking = Object2ObjectOpenHashMap<String, List<RegistryObject<RecipeDefinition>>>()
	private val input2recipesInternal = Object2ObjectOpenHashMap<String, LinkedList<RegistryObject<RecipeDefinition>>>()
	private val input2recipesBacking = Object2ObjectOpenHashMap<String, List<RegistryObject<RecipeDefinition>>>()

	val recipes: List<RegistryObject<RecipeDefinition>> = Collections.unmodifiableList(recipesInternal)
	val group2recipes: Map<String, List<RegistryObject<RecipeDefinition>>> = Collections.unmodifiableMap(group2recipesBacking)
	val output2recipes: Map<String, List<RegistryObject<RecipeDefinition>>> = Collections.unmodifiableMap(output2recipesBacking)
	val input2recipes: Map<String, List<RegistryObject<RecipeDefinition>>> = Collections.unmodifiableMap(input2recipesBacking)

	fun add(recipe: RegistryObject<RecipeDefinition>) {
		val value = recipe.value
		recipesInternal.add(recipe)

		for (group in value.groups) {
			group2recipesInternal.computeIfAbsent(group, Object2ObjectFunction { p ->
				LinkedList<RegistryObject<RecipeDefinition>>().also {
					group2recipesBacking[p as String] = Collections.unmodifiableList(it)
				}
			}).add(recipe)
		}

		output2recipesInternal.computeIfAbsent(value.output.item.name, Object2ObjectFunction { p ->
			LinkedList<RegistryObject<RecipeDefinition>>().also {
				output2recipesBacking[p as String] = Collections.unmodifiableList(it)
			}
		}).add(recipe)

		for (input in value.input) {
			input2recipesInternal.computeIfAbsent(input.item.name, Object2ObjectFunction { p ->
				LinkedList<RegistryObject<RecipeDefinition>>().also {
					input2recipesBacking[p as String] = Collections.unmodifiableList(it)
				}
			}).add(recipe)
		}
	}

	fun load(log: ILoadingLog, fileTree: Map<String, List<IStarboundFile>>, executor: ForkJoinPool): List<ForkJoinTask<*>> {
		val files = fileTree["recipe"] ?: return emptyList()

		val elements = Starbound.gson.getAdapter(JsonElement::class.java)
		val recipes = Starbound.gson.getAdapter(RecipeDefinition::class.java)

		return listOf(executor.submit {
			val line = log.line("Loading recipes...")
			val time = System.nanoTime()
			line.maxElements = files.size

			files.batch(executor) { listedFile ->
				try {
					line.text = ("Loading $listedFile")
					val json = elements.read(listedFile.jsonReader())
					val value = recipes.fromJsonTree(json)
					line.elements.incrementAndGet()
					KOptional(RegistryObject(value, json, listedFile))
				} catch (err: Throwable) {
					LOGGER.error("Loading recipe definition file $listedFile", err)
					line.elements.incrementAndGet()
					KOptional.empty()
				}
			}.forEach { add(it) }

			line.text = "Loaded recipes in ${((System.nanoTime() - time) / 1_000_000.0).toLong()}ms"
		})
	}
}
