package ru.dbotthepony.kstarbound

import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.internal.bind.JsonTreeReader
import com.google.gson.stream.JsonReader
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.api.IStarboundFile
import ru.dbotthepony.kstarbound.defs.Json2Function
import ru.dbotthepony.kstarbound.defs.JsonFunction
import ru.dbotthepony.kstarbound.defs.Species
import ru.dbotthepony.kstarbound.defs.StatusEffectDefinition
import ru.dbotthepony.kstarbound.defs.item.TreasurePoolDefinition
import ru.dbotthepony.kstarbound.defs.item.api.IItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.BackArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.ChestArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.CurrencyItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.FlashlightDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.HarvestingToolPrototype
import ru.dbotthepony.kstarbound.defs.item.impl.HeadArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.ItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.LegsArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.LiquidItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.MaterialItemDefinition
import ru.dbotthepony.kstarbound.defs.monster.MonsterSkillDefinition
import ru.dbotthepony.kstarbound.defs.monster.MonsterTypeDefinition
import ru.dbotthepony.kstarbound.defs.npc.NpcTypeDefinition
import ru.dbotthepony.kstarbound.defs.npc.TenantDefinition
import ru.dbotthepony.kstarbound.defs.`object`.ObjectDefinition
import ru.dbotthepony.kstarbound.defs.particle.ParticleDefinition
import ru.dbotthepony.kstarbound.defs.player.TechDefinition
import ru.dbotthepony.kstarbound.defs.projectile.ProjectileDefinition
import ru.dbotthepony.kstarbound.defs.quest.QuestTemplate
import ru.dbotthepony.kstarbound.defs.tile.LiquidDefinition
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kstarbound.util.KOptional
import java.util.concurrent.Callable
import java.util.concurrent.ExecutorService
import java.util.concurrent.ForkJoinPool
import java.util.concurrent.ForkJoinTask
import java.util.concurrent.Future

object Registries {
	private val LOGGER = LogManager.getLogger()

	val tiles = ObjectRegistry("tiles", TileDefinition::materialName, TileDefinition::materialId)
	val tileModifiers = ObjectRegistry("tile modifiers", MaterialModifier::modName, MaterialModifier::modId)
	val liquid = ObjectRegistry("liquid", LiquidDefinition::name, LiquidDefinition::liquidId)
	val species = ObjectRegistry("species", Species::kind)
	val statusEffects = ObjectRegistry("status effects", StatusEffectDefinition::name)
	val particles = ObjectRegistry("particles", ParticleDefinition::kind)
	val items = ObjectRegistry("items", IItemDefinition::itemName)
	val questTemplates = ObjectRegistry("quest templates", QuestTemplate::id)
	val techs = ObjectRegistry("techs", TechDefinition::name)
	val jsonFunctions = ObjectRegistry<JsonFunction>("json functions")
	val json2Functions = ObjectRegistry<Json2Function>("json 2functions")
	val npcTypes = ObjectRegistry("npc types", NpcTypeDefinition::type)
	val projectiles = ObjectRegistry("projectiles", ProjectileDefinition::projectileName)
	val tenants = ObjectRegistry("tenants", TenantDefinition::name)
	val treasurePools = ObjectRegistry("treasure pools", TreasurePoolDefinition::name)
	val monsterSkills = ObjectRegistry("monster skills", MonsterSkillDefinition::name)
	val monsterTypes = ObjectRegistry("monster types", MonsterTypeDefinition::type)
	val worldObjects = ObjectRegistry("objects", ObjectDefinition::objectName)

	private fun loadStage(
		log: ILoadingLog,
		loader: (ILoadingLog.ILine) -> Unit,
		name: String,
	) {
		if (Starbound.terminateLoading)
			return

		val time = System.currentTimeMillis()
		val line = log.line("Loading $name...".also(LOGGER::info))
		loader(line)
		line.text = ("Loaded $name in ${System.currentTimeMillis() - time}ms".also(LOGGER::info))
	}

	private inline fun <reified T : Any> loadStage(
		log: ILoadingLog,
		executor: ForkJoinPool,
		registry: ObjectRegistry<T>,
		files: List<IStarboundFile>,
		name: String = registry.name
	) {
		val adapter = Starbound.gson.getAdapter(T::class.java)
		val elementAdapter = Starbound.gson.getAdapter(JsonElement::class.java)

		loadStage(log, loader = {
			it.maxElements = files.size

			files.batch(executor) { listedFile ->
				try {
					it.text = "Loading $listedFile"

					val result = AssetPathStack(listedFile.computeDirectory()) {
						val elem = elementAdapter.read(listedFile.jsonReader())
						RegistryObject(adapter.fromJsonTree(elem), elem, listedFile)
					}

					it.elements.incrementAndGet()
					KOptional(result)
				} catch (err: Throwable) {
					LOGGER.error("Loading ${registry.name} definition file $listedFile", err)
					it.elements.incrementAndGet()
					KOptional.empty()
				}
			}.forEach { registry.add(it) }
		}, name)
	}

	fun load(log: ILoadingLog, fileTree: Map<String, List<IStarboundFile>>, executor: ForkJoinPool): List<ForkJoinTask<*>> {
		val tasks = ArrayList<ForkJoinTask<*>>()

		tasks.addAll(loadItemDefinitions(log, fileTree, executor))

		tasks.add(executor.submit { loadStage(log, { loadJsonFunctions(it, fileTree["functions"] ?: listOf()) }, "json functions") })
		tasks.add(executor.submit { loadStage(log, { loadJson2Functions(it, fileTree["2functions"] ?: listOf()) }, "json 2functions") })
		tasks.add(executor.submit { loadStage(log, { loadTreasurePools(it, fileTree["treasurepools"] ?: listOf()) }, "treasure pools") })

		tasks.add(executor.submit { loadStage(log, executor, tiles, fileTree["material"] ?: listOf()) })
		tasks.add(executor.submit { loadStage(log, executor, tileModifiers, fileTree["matmod"] ?: listOf()) })
		tasks.add(executor.submit { loadStage(log, executor, liquid, fileTree["liquid"] ?: listOf()) })
		tasks.add(executor.submit { loadStage(log, executor, worldObjects, fileTree["object"] ?: listOf()) })
		tasks.add(executor.submit { loadStage(log, executor, statusEffects, fileTree["statuseffect"] ?: listOf()) })
		tasks.add(executor.submit { loadStage(log, executor, species, fileTree["species"] ?: listOf()) })
		tasks.add(executor.submit { loadStage(log, executor, particles, fileTree["particle"] ?: listOf()) })
		tasks.add(executor.submit { loadStage(log, executor, questTemplates, fileTree["questtemplate"] ?: listOf()) })
		tasks.add(executor.submit { loadStage(log, executor, techs, fileTree["tech"] ?: listOf()) })
		tasks.add(executor.submit { loadStage(log, executor, npcTypes, fileTree["npctype"] ?: listOf()) })
		// tasks.add(executor.submit { loadStage(log, executor, projectiles, ext2files["projectile"] ?: listOf()) })
		// tasks.add(executor.submit { loadStage(log, executor, tenants, ext2files["tenant"] ?: listOf()) })
		tasks.add(executor.submit { loadStage(log, executor, monsterSkills, fileTree["monsterskill"] ?: listOf()) })
		// tasks.add(executor.submit { loadStage(log, _monsterTypes, ext2files["monstertype"] ?: listOf()) })

		return tasks
	}

	private fun loadItemDefinitions(log: ILoadingLog, files: Map<String, Collection<IStarboundFile>>, executor: ForkJoinPool): List<ForkJoinTask<*>> {
		val fileMap = mapOf(
			"item" to ItemDefinition::class.java,
			"currency" to CurrencyItemDefinition::class.java,
			"liqitem" to LiquidItemDefinition::class.java,
			"matitem" to MaterialItemDefinition::class.java,
			"flashlight" to FlashlightDefinition::class.java,
			"harvestingtool" to HarvestingToolPrototype::class.java,
			"head" to HeadArmorItemDefinition::class.java,
			"chest" to ChestArmorItemDefinition::class.java,
			"legs" to LegsArmorItemDefinition::class.java,
			"back" to BackArmorItemDefinition::class.java,
		)

		val tasks = ArrayList<ForkJoinTask<*>>()
		val objects = Starbound.gson.getAdapter(JsonObject::class.java)

		for ((ext, clazz) in fileMap) {
			val fileList = files[ext] ?: continue
			val adapter = Starbound.gson.getAdapter(clazz)

			tasks.add(executor.submit {
				val line = log.line("Loading items '$ext'")
				line.maxElements = fileList.size
				val time = System.nanoTime()

				fileList.batch(executor) { listedFile ->
					try {
						line.text = "Loading $listedFile"
						val json = objects.read(listedFile.jsonReader())
						val def = AssetPathStack(listedFile.computeDirectory()) { adapter.fromJsonTree(json) }
						line.elements.incrementAndGet()
						KOptional(RegistryObject(def, json, listedFile))
					} catch (err: Throwable) {
						LOGGER.error("Loading item definition file $listedFile", err)
						line.elements.incrementAndGet()
						KOptional.empty()
					}
				}.forEach { items.add(it) }

				line.text = "Loaded items '$ext' in ${((System.nanoTime() - time) / 1_000_000.0).toLong()}ms"
			})
		}

		return tasks
	}

	private fun loadJsonFunctions(line: ILoadingLog.ILine, files: Collection<IStarboundFile>) {
		for (listedFile in files) {
			line.text = ("Loading $listedFile")

			try {
				val json = Starbound.gson.getAdapter(JsonObject::class.java).read(JsonReader(listedFile.reader()).also { it.isLenient = true })

				for ((k, v) in json.entrySet()) {
					try {
						line.text = ("Loading $k from $listedFile")
						val fn = Starbound.gson.fromJson<JsonFunction>(JsonTreeReader(v), JsonFunction::class.java)
						jsonFunctions.add(fn, v, listedFile, k)
					} catch (err: Exception) {
						LOGGER.error("Loading json function definition $k from file $listedFile", err)
					}
				}
			} catch (err: Exception) {
				LOGGER.error("Loading json function definition $listedFile", err)
			}

			if (Starbound.terminateLoading) {
				return
			}
		}
	}

	private fun loadJson2Functions(line: ILoadingLog.ILine, files: Collection<IStarboundFile>) {
		for (listedFile in files) {
			line.text = ("Loading $listedFile")

			try {
				val json = Starbound.gson.getAdapter(JsonObject::class.java).read(JsonReader(listedFile.reader()).also { it.isLenient = true })

				for ((k, v) in json.entrySet()) {
					try {
						line.text = ("Loading $k from $listedFile")
						val fn = Starbound.gson.fromJson<Json2Function>(JsonTreeReader(v), Json2Function::class.java)
						json2Functions.add(fn, v, listedFile, k)
					} catch (err: Throwable) {
						LOGGER.error("Loading json 2function definition $k from file $listedFile", err)
					}
				}
			} catch (err: Exception) {
				LOGGER.error("Loading json 2function definition $listedFile", err)
			}

			if (Starbound.terminateLoading) {
				return
			}
		}
	}

	private fun loadTreasurePools(line: ILoadingLog.ILine, files: Collection<IStarboundFile>) {
		for (listedFile in files) {
			line.text = ("Loading $listedFile")

			try {
				val json = Starbound.gson.getAdapter(JsonObject::class.java).read(JsonReader(listedFile.reader()).also { it.isLenient = true })

				for ((k, v) in json.entrySet()) {
					try {
						line.text = ("Loading $k from $listedFile")
						val result = Starbound.gson.fromJson<TreasurePoolDefinition>(JsonTreeReader(v), TreasurePoolDefinition::class.java)
						result.name = k
						treasurePools.add(result, v, listedFile)
					} catch (err: Throwable) {
						LOGGER.error("Loading treasure pool definition $k from file $listedFile", err)
					}
				}
			} catch (err: Exception) {
				LOGGER.error("Loading treasure pool definition $listedFile", err)
			}

			if (Starbound.terminateLoading) {
				return
			}
		}
	}

}
