package ru.dbotthepony.kstarbound.client.render

import org.lwjgl.opengl.GL45.*
import ru.dbotthepony.kbox2d.api.IDebugDraw
import ru.dbotthepony.kbox2d.api.Transform
import ru.dbotthepony.kstarbound.PIXELS_IN_STARBOUND_UNIT
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.gl.vertex.GeometryType
import ru.dbotthepony.kstarbound.client.gl.vertex.StreamVertexBuilder
import ru.dbotthepony.kstarbound.client.gl.vertex.VertexAttributes
import ru.dbotthepony.kvector.arrays.Matrix3f
import ru.dbotthepony.kvector.vector.RGBAColor
import ru.dbotthepony.kvector.vector.Vector2d
import kotlin.math.cos
import kotlin.math.sin

class Box2DRenderer : IDebugDraw {
	val state = StarboundClient.current()

	override var drawShapes: Boolean = false
	override var drawJoints: Boolean = false
	override var drawAABB: Boolean = false
	override var drawTreeAABB: Boolean = false
	override var drawPairs: Boolean = false
	override var drawCenterOfMess: Boolean = false

	private val identity = Matrix3f.identity()

	private val lines = StreamVertexBuilder(VertexAttributes.POSITION_COLOR, GeometryType.LINES)
	private val triangles = StreamVertexBuilder(VertexAttributes.POSITION_COLOR, GeometryType.TRIANGLES)

	fun begin() {
		lines.builder.begin()
		triangles.builder.begin()
	}

	fun render() {
		if (lines.builder.isEmpty() && triangles.builder.isEmpty()) return

		lines.upload(GL_STREAM_DRAW)
		triangles.upload(GL_STREAM_DRAW)

		state.programs.positionColor.use()
		state.programs.positionColor.colorMultiplier = RGBAColor.WHITE
		state.programs.positionColor.modelMatrix = identity

		lines.draw(GL_LINES)
		triangles.draw(GL_TRIANGLES)
	}

	override fun drawPolygon(vertices: List<Vector2d>, color: RGBAColor) {
		require(vertices.size > 1) { "Vertex list had only ${vertices.size} namings in it" }

		if (!vertices.any { state.viewportRectangle.isInside(it) }) return

		for (i in vertices.indices) {
			val current = vertices[i]
			val next = vertices[(i + 1) % vertices.size]
			lines.builder.vertex(state.stack.last(), current.x.toFloat(), current.y.toFloat()).color(color)
			lines.builder.vertex(state.stack.last(), next.x.toFloat(), next.y.toFloat()).color(color)
		}
	}

	private fun drawSolid(vertices: List<Vector2d>, color: RGBAColor) {
		require(vertices.size >= 3) { "Vertex list had only ${vertices.size} namings in it" }

		if (!vertices.any { state.viewportRectangle.isInside(it) }) return

		val zero = vertices[0]

		for (i in 1 until vertices.size) {
			val current = vertices[i]
			val next = vertices[(i + 1) % vertices.size]
			triangles.builder.vertex(state.stack.last(), zero.x.toFloat(), zero.y.toFloat()).color(color)
			triangles.builder.vertex(state.stack.last(), current.x.toFloat(), current.y.toFloat()).color(color)
			triangles.builder.vertex(state.stack.last(), next.x.toFloat(), next.y.toFloat()).color(color)
		}
	}

	override fun drawSolidPolygon(vertices: List<Vector2d>, color: RGBAColor) {
		drawSolid(vertices, color.copy(alpha = 0.5f))
		drawPolygon(vertices, color)
	}

	override fun drawCircle(center: Vector2d, radius: Double, color: RGBAColor) {
		val vertexList = ArrayList<Vector2d>()

		for (i in 0 until 360 step 15) {
			val rad = Math.toRadians(i.toDouble())
			val c = cos(rad)
			val s = sin(rad)

			vertexList.add(Vector2d(
				center.x + c * radius,
				center.y + s * radius
			))
		}

		drawPolygon(vertexList, color)
	}

	override fun drawSolidCircle(center: Vector2d, radius: Double, axis: Vector2d, color: RGBAColor) {
		val vertexList = ArrayList<Vector2d>()

		for (i in 0 until 360 step 15) {
			val rad = Math.toRadians(i.toDouble())
			val c = cos(rad)
			val s = sin(rad)

			vertexList.add(Vector2d(
				center.x + c * radius,
				center.y + s * radius
			))
		}

		drawSolidPolygon(vertexList, color.copy(alpha = 0.5f))
		drawPolygon(vertexList, color)
		drawPolygon(listOf(center, center + axis * radius), color)
	}

	override fun drawSegment(p1: Vector2d, p2: Vector2d, color: RGBAColor) {
		drawPolygon(listOf(p1, p2), color)
	}

	override fun drawTransform(xf: Transform) {
		TODO("Not yet implemented")
	}

	override fun drawPoint(p: Vector2d, size: Double, color: RGBAColor) {
		drawSolid(listOf(
			Vector2d(x = p.x - size / (PIXELS_IN_STARBOUND_UNIT * 2.0), y = p.y - size / (PIXELS_IN_STARBOUND_UNIT * 2.0)),
			Vector2d(x = p.x + size / (PIXELS_IN_STARBOUND_UNIT * 2.0), y = p.y - size / (PIXELS_IN_STARBOUND_UNIT * 2.0)),
			Vector2d(x = p.x + size / (PIXELS_IN_STARBOUND_UNIT * 2.0), y = p.y + size / (PIXELS_IN_STARBOUND_UNIT * 2.0)),
			Vector2d(x = p.x - size / (PIXELS_IN_STARBOUND_UNIT * 2.0), y = p.y + size / (PIXELS_IN_STARBOUND_UNIT * 2.0)),
		), color)
	}
}
