package ru.dbotthepony.kstarbound.player

import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import it.unimi.dsi.fastutil.objects.Object2LongOpenHashMap
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.RegistryObject
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.player.TechDefinition
import ru.dbotthepony.kstarbound.lua.LuaState
import ru.dbotthepony.kstarbound.lua.loadInternalScript
import ru.dbotthepony.kstarbound.util.ItemStack
import ru.dbotthepony.kstarbound.util.immutableMap
import java.util.*
import kotlin.collections.ArrayList

/**
 * Персонаж - как он есть.
 *
 * [Avatar] реализует Lua интерфейс `player`.
 */
class Avatar(val uniqueId: UUID) {
	enum class EssentialSlot {
		BEAM_AXE,
		WIRE_TOOL,
		PAINT_TOOL,
		INSPECTION_TOOL;
	}

	enum class EquipmentSlot {
		HEAD,
		CHEST,
		LEGS,
		BACK,
		HEAD_COSMETIC,
		CHEST_COSMETIC,
		LEGS_COSMETIC,
		BACK_COSMETIC;
	}

	private val essentialSlots = EnumMap<EssentialSlot, ItemStack>(EssentialSlot::class.java)
	private val equipmentSlots = EnumMap<EquipmentSlot, ItemStack>(EquipmentSlot::class.java)
	private val bags = ArrayList<AvatarBag>()
	private val quests = Object2ObjectOpenHashMap<String, QuestInstance>()

	var cursorItem = ItemStack.EMPTY

	private val availableTechs = ObjectOpenHashSet<RegistryObject<TechDefinition>>()
	private val enabledTechs = ObjectOpenHashSet<RegistryObject<TechDefinition>>()
	private val equippedTechs = Object2ObjectOpenHashMap<String, RegistryObject<TechDefinition>>()

	private val knownBlueprints = ObjectOpenHashSet<ItemStack>()
	// С подписью NEW
	private val newBlueprints = ObjectOpenHashSet<ItemStack>()

	private val currencies = Object2LongOpenHashMap<String>()

	/**
	 * Teaches the player any recipes which can be used to craft the specified item.
	 */
	fun giveBlueprint(name: String): Boolean {
		val item = Starbound.item(name).conciseToNull() ?: return false

		if (knownBlueprints.add(item)) {
			newBlueprints.add(item)
			return true
		}

		return false
	}

	/**
	 * Returns `true` if the player knows one or more recipes to create the specified item and `false` otherwise.
	 */
	fun blueprintKnown(name: String): Boolean {
		return (Starbound.item(name).conciseToNull() ?: return false) in knownBlueprints
	}

	/**
	 * Returns `true` if the player knows one or more recipes to create the specified item and `false` otherwise.
	 */
	private fun blueprintKnown(name: JsonElement): Boolean {
		if (name is JsonPrimitive) {
			return (Starbound.item(name.asString).conciseToNull() ?: return false) in knownBlueprints
		} else if (name is JsonObject) {
			return (Starbound.item(name).conciseToNull() ?: return false) in knownBlueprints
		} else {
			return false
		}
	}

	/**
	 * Teaches the player any recipes which can be used to craft the specified item.
	 */
	private fun giveBlueprint(name: JsonElement): Boolean {
		val item: ItemStack

		if (name is JsonPrimitive) {
			item = Starbound.item(name.asString).conciseToNull() ?: return false
		} else if (name is JsonObject) {
			item = Starbound.item(name).conciseToNull() ?: return false
		} else {
			return false
		}

		if (knownBlueprints.add(item)) {
			newBlueprints.add(item)
			return true
		}

		return false
	}

	/**
	 * Adds the specified tech to the player's list of available (unlockable) techs.
	 */
	fun makeTechAvailable(name: String): Boolean {
		return availableTechs.add(Registries.techs[name] ?: return false)
	}

	/**
	 * Removes the specified tech from player's list of available (unlockable) techs.
	 */
	fun makeTechUnavailable(name: String): Boolean {
		val tech = Registries.techs[name] ?: return false

		if (availableTechs.remove(tech)) {
			enabledTechs.remove(tech)
			equippedTechs.remove(tech.value.type)
			return true
		}

		return false
	}

	/**
	 * Unlocks the specified tech, allowing it to be equipped through the tech GUI.
	 */
	fun enableTech(name: String): Boolean {
		val tech = Registries.techs[name] ?: return false
		availableTechs.add(tech)
		return enabledTechs.add(tech)
	}

	/**
	 * Equips the specified tech.
	 */
	fun equipTech(name: String): Boolean {
		val tech = Registries.techs[name] ?: return false
		availableTechs.add(tech)
		enabledTechs.add(tech)
		return equippedTechs.put(tech.value.type, tech) != tech
	}

	/**
	 * Unequips the specified tech.
	 */
	fun unequipTech(name: String): Boolean {
		val tech = Registries.techs[name] ?: return false
		return equippedTechs.remove(tech.value.type) == tech
	}

	/**
	 * Returns the player's current total reserves of the specified currency.
	 */
	fun currency(name: String): Long {
		return currencies.getLong(name)
	}

	/**
	 * Increases the player's reserve of the specified currency by the specified amount.
	 */
	fun addCurrency(name: String, amount: Long) {
		check(amount >= 0L) { "Negative amount of currency: $amount (currency: $name)" }
		currencies.computeLong(name) { key, old -> (old ?: 0L) + amount }
	}

	/**
	 * Attempts to consume the specified amount of the specified currency and returns `true` if successful and `false` otherwise.
	 */
	fun consumeCurrency(name: String, amount: Long): Boolean {
		check(amount >= 0L) { "Negative amount of currency: $amount (currency: $name)" }
		val current = currencies.getLong(name)

		if (current - amount >= 0L) {
			currencies[name] = current - amount
			return true
		}

		return false
	}

	/**
	 * Triggers an immediate cleanup of the player's inventory, removing item stacks with 0 quantity. May rarely be required in special cases of making several sequential modifications to the player's inventory within a single tick.
	 */
	fun cleanupItems() {
		TODO()
	}

	/**
	 * Adds the specified item to the player's inventory.
	 */
	fun giveItem(descriptor: ItemStack) {
		TODO()
	}

	/**
	 * Returns `true` if the player's inventory contains an item matching the specified descriptor and `false` otherwise. If exactMatch is `true` then parameters as well as item name must match.
	 */
	fun hasItem(descriptor: ItemStack, exactMatch: Boolean = false): Boolean {
		return false
	}

	/**
	 * Returns the total number of items in the player's inventory matching the specified descriptor. If exactMatch is `true` then parameters as well as item name must match.
	 */
	fun hasCountOfItem(descriptor: ItemStack, exactMatch: Boolean = false): Long {
		return 0L
	}

	/**
	 * Attempts to consume the specified item from the player's inventory and returns the item consumed if successful. If consumePartial is `true`, matching stacks totalling fewer items than the requested count may be consumed, otherwise the operation will only be performed if the full count can be consumed. If exactMatch is `true` then parameters as well as item name must match.
	 */
	fun consumeItem(descriptor: ItemStack, allowPartial: Boolean = false, exactMatch: Boolean = false): ItemStack {
		return ItemStack.EMPTY
	}

	fun inventoryTags(): Map<String, Long> {
		return mapOf()
	}

	fun itemsWithTag(): List<ItemStack> {
		return listOf()
	}

	fun consumeTaggedItem(tag: String): Long {
		return 0L
	}

	fun hasItemWithParameter(name: String, value: JsonElement): Boolean {
		return false
	}

	fun consumeItemWithParameter(name: String, value: JsonElement, count: Long): Long {
		return 0L
	}

	fun getItemWithParameter(name: String, value: JsonElement): ItemStack {
		return ItemStack.EMPTY
	}

	var primaryHandItem: ItemStack? = null
	var altHandItem: ItemStack? = null

	fun essentialItem(slotName: EssentialSlot): ItemStack? {
		return essentialSlots[slotName]?.conciseToNull()
	}

	fun giveEssentialItem(slotName: EssentialSlot, item: ItemStack) {

	}

	fun removeEssentialItem(slotName: EssentialSlot) {

	}

	fun equippedItem(slotName: EquipmentSlot): ItemStack {
		return equipmentSlots[slotName] ?: ItemStack.EMPTY
	}

	fun setEquippedItem(slotName: EquipmentSlot, item: ItemStack) {

	}

	fun addQuest(quest: QuestInstance): QuestInstance? {
		check(quest.avatar === this) { "$quest does not belong to $this" }
		return quests.put(quest.id, quest)
	}

	private fun startQuest(value: JsonElement, serverID: String?, worldID: String?): String {
		if (value is JsonPrimitive) {
			val quest = QuestInstance(this, descriptor = QuestDescriptor(value.asString), serverID = serverID?.let(UUID::fromString), worldID = worldID)
			addQuest(quest)
			return quest.id
		} else if (value is JsonObject) {
			val seed = value["seed"]?.asLong ?: QuestDescriptor.makeSeed()
			val questId = value["questId"]?.asString ?: throw IllegalArgumentException("Invalid 'questId' in quest descriptor")
			val templateId = value["templateId"]?.asString ?: questId
			val params = value["parameters"] as? JsonObject ?: JsonObject()
			val quest = QuestInstance(this, descriptor = QuestDescriptor(questId, templateId, seed, params), serverID = serverID?.let(UUID::fromString), worldID = worldID)
			addQuest(quest)
			return quest.id
		} else {
			throw IllegalArgumentException("Invalid quest descriptor: $value")
		}
	}

	fun pushLuaAPI(lua: LuaState) {
		lua.load(playerLua)

		lua.pushTable()

		lua.setTableValue("uniqueId", uniqueId.toString())
		lua.setTableValue("species", "human")
		lua.setTableValue("gender", "male")

		lua.call(numArgs = 1)

		lua.loadGlobal("player")
		lua.setTableClosure("giveBlueprint", this) { giveBlueprint(it.getValue()) }
		lua.setTableFunction("blueprintKnown", this) { it.lua.push(blueprintKnown(it.getValue())); 1 }
		lua.setTableClosure("makeTechAvailable", this) { makeTechAvailable(it.getString()) }
		lua.setTableClosure("makeTechUnavailable", this) { makeTechUnavailable(it.getString()) }
		lua.setTableClosure("enableTech", this) { enableTech(it.getString()) }
		lua.setTableClosure("equipTech", this) { equipTech(it.getString()) }
		lua.setTableClosure("unequipTech", this) { unequipTech(it.getString()) }
		lua.setTableFunction("availableTechs", this) { it.lua.pushStrings(availableTechs.stream().map { it.value.name }.iterator()); 1 }
		lua.setTableFunction("enabledTechs", this) { it.lua.pushStrings(enabledTechs.stream().map { it.value.name }.iterator()); 1 }
		lua.setTableFunction("equippedTech", this) { it.lua.push(equippedTechs[it.getString()]?.value?.name); 1 }
		lua.setTableFunction("currency", this) { it.lua.push(currency(it.getString())); 1 }
		lua.setTableClosure("addCurrency", this) { addCurrency(it.getString(), it.getLong()) }
		lua.setTableFunction("consumeCurrency", this) { it.lua.push(consumeCurrency(it.getString(), it.getLong())); 1 }
		lua.setTableClosure("cleanupItems", this) { cleanupItems() }
		lua.setTableClosure("giveItem", this) { giveItem(Starbound.item(it.getValue())) }

		lua.setTableFunction("hasItem", this) {
			it.lua.push(hasItem(Starbound.item(it.getValue()), it.getBoolOrNull() ?: false))
			1
		}

		lua.setTableFunction("hasCountOfItem", this) {
			it.lua.push(hasCountOfItem(Starbound.item(it.getValue()), it.getBoolOrNull() ?: false))
			1
		}

		lua.setTableFunction("consumeItem", this) {
			it.lua.push(consumeItem(Starbound.item(it.getValue()), it.getBoolOrNull() ?: false, it.getBoolOrNull() ?: false).toJson())
			1
		}

		lua.setTableFunction("inventoryTags", this) {
			val mapping = inventoryTags()

			it.lua.pushTable(hashSize = mapping.size)

			for ((k, v) in mapping) {
				it.lua.setTableValue(k, v)
			}

			1
		}

		lua.setTableFunction("itemsWithTag", this) {
			val mapping = itemsWithTag()

			it.lua.pushTable(arraySize = mapping.size)

			for ((k, v) in mapping.withIndex()) {
				it.lua.setTableValue(k, v.toJson())
			}

			1
		}

		lua.setTableFunction("consumeTaggedItem", this) {
			it.lua.push(consumeTaggedItem(it.getString()))
			1
		}

		lua.setTableFunction("hasItemWithParameter", this) {
			it.lua.push(hasItemWithParameter(it.getString(), it.getValue()))
			1
		}

		lua.setTableFunction("consumeItemWithParameter", this) {
			it.lua.push(consumeItemWithParameter(it.getString(), it.getValue(), it.getLong()))
			1
		}

		lua.setTableFunction("getItemWithParameter", this) {
			it.lua.push(getItemWithParameter(it.getString(), it.getValue()).toJson())
			1
		}

		lua.setTableFunction("primaryHandItem", this) {
			it.lua.push(primaryHandItem?.toJson())
			1
		}

		lua.setTableFunction("altHandItem", this) {
			it.lua.push(altHandItem?.toJson())
			1
		}

		lua.setTableFunction("primaryHandItemTags", this) {
			val tags = primaryHandItem?.item?.value?.itemTags
			if (tags != null) it.lua.pushStrings(tags) else it.lua.push()
			1
		}

		lua.setTableFunction("altHandItemTags", this) {
			val tags = altHandItem?.item?.value?.itemTags
			if (tags != null) it.lua.pushStrings(tags) else it.lua.push()
			1
		}

		lua.setTableFunction("essentialItem", this) {
			val name = it.getString()
			val slot = essentialSlotsMap[name] ?: throw IllegalArgumentException("Invalid slot '$name'")
			it.lua.push(essentialItem(slot)?.toJson())
			1
		}

		lua.setTableClosure("giveEssentialItem", this) {
			val name = it.getString()
			val item = Starbound.item(it.getValue())
			val slot = essentialSlotsMap[name] ?: throw IllegalArgumentException("Invalid slot '$name'")
			giveEssentialItem(slot, item)
		}

		lua.setTableClosure("removeEssentialItem", this) {
			val name = it.getString()
			val slot = essentialSlotsMap[name] ?: throw IllegalArgumentException("Invalid slot '$name'")
			removeEssentialItem(slot)
		}

		lua.setTableFunction("equippedItem", this) {
			val name = it.getString()
			val slot = equipmentSlotsMap[name] ?: throw IllegalArgumentException("Invalid slot '$name'")
			it.lua.push(equippedItem(slot).toJson())
			1
		}

		lua.setTableClosure("setEquippedItem", this) {
			val name = it.getString()
			val item = Starbound.item(it.getValue())
			val slot = equipmentSlotsMap[name] ?: throw IllegalArgumentException("Invalid slot '$name'")
			setEquippedItem(slot, item)
		}

		lua.setTableFunction("swapSlotItem", this) {
			it.lua.push(cursorItem.toJson())
			1
		}

		lua.setTableClosure("setSwapSlotItem", this) {
			cursorItem = Starbound.item(it.getValue())
		}

		lua.setTableFunction("startQuest", this) {
			it.lua.push(startQuest(it.getValue(), it.getStringOrNull(), it.getStringOrNull()))
			1
		}

		lua.setTableFunction("hasQuest", this) {
			it.lua.push(quests[it.getString()] != null)
			1
		}

		lua.setTableFunction("hasCompletedQuest", this) {
			val quest = quests[it.getString()]
			it.lua.push(quest != null && quest.state == QuestInstance.State.COMPLETE)
			1
		}

		lua.pop()
	}

	companion object {
		private val playerLua by lazy { loadInternalScript("player") }

		private val essentialSlotsMap = immutableMap<String, EssentialSlot> {
			put("beamaxe", EssentialSlot.BEAM_AXE)
			put("inspectiontool", EssentialSlot.INSPECTION_TOOL)
			put("wiretool", EssentialSlot.WIRE_TOOL)
			put("painttool", EssentialSlot.PAINT_TOOL)
		}

		private val equipmentSlotsMap = immutableMap<String, EquipmentSlot> {
			put("head", EquipmentSlot.HEAD)
			put("chest", EquipmentSlot.CHEST)
			put("legs", EquipmentSlot.LEGS)
			put("back", EquipmentSlot.BACK)
			put("headCosmetic", EquipmentSlot.HEAD_COSMETIC)
			put("chestCosmetic", EquipmentSlot.CHEST_COSMETIC)
			put("legsCosmetic", EquipmentSlot.LEGS_COSMETIC)
			put("backCosmetic", EquipmentSlot.BACK_COSMETIC)
		}
	}
}
