package ru.dbotthepony.kstarbound.world.entities

import com.google.common.collect.ImmutableMap
import com.google.gson.JsonObject
import com.google.gson.TypeAdapter
import com.google.gson.reflect.TypeToken
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.RegistryObject
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.client.world.ClientWorld
import ru.dbotthepony.kstarbound.defs.Drawable
import ru.dbotthepony.kstarbound.defs.JsonDriven
import ru.dbotthepony.kstarbound.defs.image.SpriteReference
import ru.dbotthepony.kstarbound.defs.`object`.ObjectDefinition
import ru.dbotthepony.kstarbound.defs.`object`.ObjectOrientation
import ru.dbotthepony.kstarbound.util.get
import ru.dbotthepony.kstarbound.util.set
import ru.dbotthepony.kstarbound.world.Side
import ru.dbotthepony.kstarbound.world.LightCalculator
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.api.TileColor
import ru.dbotthepony.kvector.vector.RGBAColor
import ru.dbotthepony.kvector.vector.Vector2i

open class WorldObject(
	val world: World<*, *>,
	val prototype: RegistryObject<ObjectDefinition>,
	val pos: Vector2i,
) : JsonDriven(prototype.file.computeDirectory()) {
	constructor(world: World<*, *>, data: JsonObject) : this(
		world,
		Registries.worldObjects[data["name"]?.asString ?: throw IllegalArgumentException("Missing object name")] ?: throw IllegalArgumentException("No such object defined for '${data["name"]}'"),
		data.get("tilePosition", vectors) { throw IllegalArgumentException("No tilePosition was present in saved data") }
	) {
		direction = data.get("direction", directions) { Side.LEFT }
		orientationIndex = data.get("orientationIndex", -1)
		interactive = data.get("interactive", false)

		data["uniqueId"]?.let {
			if (!it.isJsonNull)
				uniqueId = it.asString
		}

		for ((k, v) in data.get("parameters") { JsonObject() }.entrySet()) {
			properties[k] = v.deepCopy()
		}
	}

	//
	// internal runtime properties
	//
	val clientWorld get() = world as ClientWorld
	val orientations = prototype.value.orientations
	protected val renderParamLocations = Object2ObjectOpenHashMap<String, () -> String?>()
	private var frame = 0
		set(value) {
			if (field != value) {
				field = value
				drawablesCache.invalidate()
			}
		}

	private var frameTimer = 0.0
	val flickerPeriod = prototype.value.flickerPeriod?.copy()

	var isRemoved = false
		private set
	var isSpawned = false
		private set

	//
	// top level properties
	//
	var uniqueId: String? = null
	var interactive = false
	var direction = Side.LEFT

	var orientationIndex = -1
		set(value) {
			if (field != value) {
				field = value
				invalidate()
			}
		}

	//
	// json driven properties
	//
	var color: TileColor by Property(TileColor.DEFAULT)

	var animationParts: ImmutableMap<String, SpriteReference> by Property()
	var imagePosition: Vector2i by Property(Vector2i.ZERO)
	var animationPosition: Vector2i by Property(Vector2i.ZERO)

	init {
		renderParamLocations["color"] = { color.lowercase }
		renderParamLocations["frame"] = { frame.toString() }
	}

	private val drawablesCache = LazyData {
		orientation?.drawables?.map { it.with(::getRenderParam) } ?: listOf()
	}

	val drawables: List<Drawable> by drawablesCache

	fun getRenderParam(key: String): String? {
		return renderParamLocations[key]?.invoke() ?: "default"
	}

	override fun invalidate(name: String) {
		super.invalidate(name)

		if (name in renderParamLocations) {
			drawablesCache.invalidate()
		}
	}

	override fun invalidate() {
		super.invalidate()
	}

	protected open fun innerSpawn() {}
	protected open fun innerRemove() {}

	fun spawn() {
		if (isSpawned) return
		isSpawned = true
		world.objects.add(this)
		innerSpawn()
		invalidate()
	}

	open fun remove() {
		if (isRemoved || !isSpawned) return
		isRemoved = true
		check(world.objects.remove(this))
		innerRemove()
	}

	open fun thinkShared() {
		flickerPeriod?.update(Starbound.TICK_TIME_ADVANCE, world.random)
	}

	open fun thinkClient() {
		val orientation = orientation

		if (orientation != null) {
			frameTimer = (frameTimer + Starbound.TICK_TIME_ADVANCE) % orientation.animationCycle
			frame = (frameTimer / orientation.animationCycle * orientation.frames).toInt()
		}
	}

	open fun thinkServer() {

	}

	val orientation: ObjectOrientation? get() {
		return orientations.getOrNull(orientationIndex)
	}

	override fun defs(): Collection<JsonObject> {
		val get = orientation
		return if (get == null) listOf(prototype.jsonObject) else listOf(get.json, prototype.jsonObject)
	}

	val lightColors: ImmutableMap<String, RGBAColor> by LazyData(listOf("lightColor", "lightColors")) {
		dataValue("lightColor")?.let { ImmutableMap.of("default", colors0.fromJsonTree(it)) }
			?: dataValue("lightColors")?.let { colors1.fromJsonTree(it) }
			?: ImmutableMap.of()
	}

	fun addLights(lightCalculator: LightCalculator, xOffset: Int, yOffset: Int) {
		var color = lightColors[color.lowercase]

		if (color != null) {
			if (flickerPeriod != null) {
				val sample = flickerPeriod.sinValue().toFloat()
				color *= sample
			}

			lightCalculator.addPointLight(pos.x - xOffset, pos.y - yOffset, color)
		}
	}

	companion object {
		private val colors1 by lazy { Starbound.gson.getAdapter(TypeToken.getParameterized(ImmutableMap::class.java, String::class.java, RGBAColor::class.java)) as TypeAdapter<ImmutableMap<String, RGBAColor>> }
		private val colors0 by lazy { Starbound.gson.getAdapter(RGBAColor::class.java) }
		private val strings by lazy { Starbound.gson.getAdapter(String::class.java) }
		private val directions by lazy { Starbound.gson.getAdapter(Side::class.java) }
		private val vectors by lazy { Starbound.gson.getAdapter(Vector2i::class.java) }
	}
}
